@@ -1,10 +1,16 @@
 from napari_plugin_engine import napari_hook_implementation
+
+from ._clustering import ClusteringWidget
+from ._dimensionality_reduction import DimensionalityReductionWidget
 from ._measure import MeasureWidget
 from ._plotter import PlotterWidget
-from ._dimensionality_reduction import DimensionalityReductionWidget
-from ._clustering import ClusteringWidget
 
 
 @napari_hook_implementation
 def napari_experimental_provide_dock_widget():
-    return [MeasureWidget, PlotterWidget, DimensionalityReductionWidget, ClusteringWidget]
+    return [
+        MeasureWidget,
+        PlotterWidget,
+        DimensionalityReductionWidget,
+        ClusteringWidget,
+    ]
